# -*- coding: utf-8 -*-
#/usr/bin/python
#
#% $Id: connect_meqtimba_dialog.py 5418 2007-07-19 16:49:13Z oms $ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from MeqGUI.GUI import meqgui
from MeqGUI.GUI.pixmaps import pixmaps

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

import os
import os.path

class ServersDialog (QDialog):
  def __init__(self,parent=None,name=None,modal=0,fl=None):
    if fl is None:
      fl = Qt.Dialog|Qt.WindowTitleHint;
    QDialog.__init__(self,parent,Qt.Dialog|Qt.WindowTitleHint);
    self.setModal(modal);
    
    self._default_path = "meqserver";
    
    self.image0 = pixmaps.trees48x48.pm();

    # self.setSizeGripEnabled(0)
    lo_dialog = QVBoxLayout(self);

    LayoutWidget = QWidget(self);
    lo_dialog.addWidget(LayoutWidget);
    LayoutWidget.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding);
    # LayoutWidget.setGeometry(QRect(10,10,472,400))
    lo_top = QVBoxLayout(LayoutWidget);

    lo_title = QHBoxLayout(None);

    self.title_icon = QLabel(LayoutWidget)
    self.title_icon.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed));
    self.title_icon.setPixmap(self.image0)
    self.title_icon.setAlignment(Qt.AlignCenter)
    lo_title.addWidget(self.title_icon)

    self.title_label = QLabel(LayoutWidget)
    self.title_label.setWordWrap(True);
    lo_title.addWidget(self.title_label)
    lo_top.addLayout(lo_title)
            
    self.gb_servers = QGroupBox(LayoutWidget);
    lo = QVBoxLayout(self.gb_servers);
    self.server_list = QTreeWidget(self.gb_servers);
    lo.addWidget(self.server_list);
    # self.server_list.header().hide();
    self.server_list.setHeaderLabels(["server","stat","dir","script"]);
    self.server_list.header().setResizeMode(0,QHeaderView.ResizeToContents);
    self.server_list.header().setResizeMode(1,QHeaderView.ResizeToContents);
    self.server_list.header().setStretchLastSection(False);
    try:
      self.server_list.setAllColumnsShowFocus(True);
    except AttributeError:
      pass;
    QObject.connect(self.server_list,SIGNAL("itemActivated(QTreeWidgetItem*,int)"),self._server_selected);
    lo_top.addWidget(self.gb_servers)

    self.gb_local = QGroupBox(LayoutWidget)
    lo = QGridLayout(self.gb_local);
    # lo.setSpacing(0);
    
    lo_start_lbl = QLabel("program:",self.gb_local);
    lo.addWidget(lo_start_lbl,0,0)
    self.start_pathname = QLineEdit(self.gb_local)
    self.start_pathname.setText(self._default_path);
    self.start_pathname.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.Fixed);
    self.start_pathname.setMinimumSize(QSize(400,0))
    lo.addWidget(self.start_pathname,0,1)

    #lo_start2.setContentsMargins(0,0,0,0);
    #lo_start_space2 = QSpacerItem(20,20,QSizePolicy.Fixed,QSizePolicy.Minimum)
    #lo_start2.addItem(lo_start_space2)
    lo_start_lbl2 = QLabel("args:",self.gb_local);
    lo.addWidget(lo_start_lbl2,1,0)
    self.start_args = QLineEdit(self.gb_local)
    lo.addWidget(self.start_args,1,1);
    
    lo_start_grp3 = QWidget(self.gb_local);
    lo.addWidget(lo_start_grp3,2,1);
    lo_start3 = QHBoxLayout(lo_start_grp3)
    #lo_start3.setContentsMargins(0,0,0,0);
    self.start_local = QPushButton(self.gb_local);
    lo.addWidget(self.start_local,2,0);
    self.start_local.setAutoDefault(1);
    self.start_local.setDefault(1);
    lo_start3.addStretch(1);
    self.start_browse = QPushButton(lo_start_grp3)
    lo_start3.addWidget(self.start_browse,0)
    self.start_default = QPushButton(lo_start_grp3)
    self.start_default.setEnabled(False);
    lo_start3.addWidget(self.start_default,0)
    QObject.connect(self.start_local,SIGNAL("clicked()"),
                    self._start_local_server_selected);

    #lo_remote_grp = QWidget(self.gb_local)
    #lo_remote = QHBoxLayout(lo_remote_grp,0,6,"lo_remote")
    #lo_remote_space = QSpacerItem(20,20,QSizePolicy.Fixed,QSizePolicy.Minimum)
    #lo_remote.addItem(lo_remote_space)

    #self.remote_host_lbl = QLabel(lo_remote_grp,"remote_host_lbl")
    #self.remote_host_lbl.setAlignment(QLabel.AlignVCenter | QLabel.AlignRight)
    #lo_remote.addWidget(self.remote_host_lbl)

    #self.remote_host = QLineEdit(lo_remote_grp,"remote_host")
    #lo_remote.addWidget(self.remote_host)

    #self.remote_port_lbl = QLabel(lo_remote_grp,"remote_port_lbl")
    #self.remote_port_lbl.setAlignment(QLabel.AlignVCenter | QLabel.AlignRight)
    #lo_remote.addWidget(self.remote_port_lbl)

    #self.remote_port = QLineEdit(lo_remote_grp,"remote_port")
    #self.remote_port.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Fixed);
    #self.remote_port.setMinimumSize(QSize(60,0))
    #lo_remote.addWidget(self.remote_port)
    #lo_connect.addWidget(lo_remote_grp)
    #lo_remote_grp.setEnabled(False);
    
    lo_top.addWidget(self.gb_local)

    lo_mainbtn = QHBoxLayout(None)
    #lo_mainbtn_space = QSpacerItem(20,20,QSizePolicy.Expanding,QSizePolicy.Minimum)
    #lo_mainbtn.addItem(lo_mainbtn_space)

    #self.btn_ok = QPushButton(LayoutWidget,"btn_ok")
    #self.btn_ok.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed,1,0,self.btn_ok.sizePolicy().hasHeightForWidth()))
    #self.btn_ok.setMinimumSize(QSize(60,0))
    #self.btn_ok.setAutoDefault(1)
    #self.btn_ok.setDefault(1)
    #lo_mainbtn.addWidget(self.btn_ok)

    lo_mainbtn.addStretch(1);
    self.btn_cancel = QPushButton(LayoutWidget)
    self.btn_cancel.setSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed);
    self.btn_cancel.setMinimumSize(QSize(60,0))
    self.btn_cancel.setAutoDefault(1)
    self.btn_cancel.setIcon(pixmaps.red_round_cross.icon());
    lo_mainbtn.addWidget(self.btn_cancel,0)
    
    #lo_space = QSpacerItem(20,20,QSizePolicy.Fixed,QSizePolicy.Expanding);
    #lo_top.addItem(lo_space)
    lo_top.addLayout(lo_mainbtn);

    self.languageChange()
    
    LayoutWidget.adjustSize();

    #LayoutWidget.resize(QSize(489,330).expandedTo(LayoutWidget.minimumSizeHint()))
    #self.resize(QSize(489,330).expandedTo(self.minimumSizeHint()))
    # self.clearWState(Qt.WState_Polished)
    
    self.connect(self.btn_cancel,SIGNAL("clicked()"),self.reject)
    
    ### my additions
    #self.connect(self.btn_remote,SIGNAL("toggled(bool)"),lo_remote_grp,SLOT("setEnabled(bool)"));
    self.connect(self.start_browse,SIGNAL("clicked()"),self.browse_kernel_dialog);
    self.connect(self.start_default,SIGNAL("clicked()"),self.reset_default_path);
    self.connect(self.start_pathname,SIGNAL("textChanged(const QString &)"),self.changed_path);


  def languageChange(self):
    self.setWindowTitle(self.__tr("Attach to a meqserver"))
    self.title_label.setText(self.__tr( \
      """<p><i>Pick a running meqserver to attach to, or else start a new one
      using one of the methods below.</i></p>
      <p><i>You can also start a meqserver process externally, e.g. from a terminal or from a debugger,
      and it should show up in the list.</i></p>""" \
      ));
    self.gb_servers.setTitle(self.__tr("Running meqservers (double-click to attach)"))
    self.gb_local.setTitle(self.__tr("Start && attach a new local meqserver"))
    # self.btn_wait.setText(self.__tr("wait for local connection"))
    self.start_browse.setText(self.__tr("Browse..."))
    self.start_default.setText(self.__tr("Reset"))
    self.start_local.setText(self.__tr("Start"))
    #self.btn_remote.setText(self.__tr("connect to remote kernel:"))
    #self.remote_host_lbl.setText(self.__tr("Host"))
    #self.remote_port_lbl.setText(self.__tr("Port"))
    #self.remote_port.setInputMask(self.__tr("#####; "))
    self.btn_cancel.setText(self.__tr("Cancel"))


  def __tr(self,s,c = None):
    return qApp.translate("ConnectMeqKernel",s,c)

  def set_default_args (self,args):
    self.start_args.setText(args);
  
  def set_default_path (self,path):
    self._default_path = path;
    self.start_pathname.setText(path);
    self.start_default.setEnabled(False);

  def reset_default_path (self):
    self.start_pathname.setText(self._default_path);
    self.start_default.setEnabled(False);
    
  def changed_path (self,path):
    self.start_default.setEnabled(bool(str(path) != self._default_path));

  def browse_kernel_dialog (self):
    try: dialog = self._browse_dialog;
    except AttributeError:
      self._browse_dialog = dialog = QFileDialog(self,"Select meqserver binary");
      dialog.resize(500,dialog.height());
      dialog.setFileMode(QFileDialog.ExistingFile);
      dialog.setViewMode(QFileDialog.Detail);
    if dialog.exec_() == QDialog.Accepted:
      self.start_pathname.setText(str(dialog.selectedFiles()[0]));
  
  def _server_selected (self,item,*dum):
    self.emit(SIGNAL("serverSelected"),item._addr,);
    
  def _start_local_server_selected (self):
    pathname = str(self.start_pathname.text());
    args = str(self.start_args.text());
    self.emit(SIGNAL("startKernel"),pathname,args);
  
  def attach_to_server (self,addr):
    for index in range(self.server_list.topLevelItemCount()):
      item = self.server_list.topLevelItem(index);
      if item._addr == addr:
        item.setIcon(0,pixmaps.blue_round_rightarrow.icon());
      else:
        item.setIcon(0,QIcon());
    
  def update_server_state (self,server,staterec=None):
    # find matching item
    for index in range(self.server_list.topLevelItemCount()):
      item = self.server_list.topLevelItem(index);
      if item._addr == server.addr:
        break;
    else:
      item = index = None;
    # disconnected server: remove from list
    if server.state is None:
      if index is not None:
        self.server_list.takeTopLevelItem(index);
    else:
      # connected server: add to list and/or change state
      if item is None:
        item = QTreeWidgetItem(self.server_list);
        setattr(item,'_addr',server.addr);
      # change fields
      name = server.session_name or server.addr;
      if server.remote and server.host:
        name += ' @'+server.host;
      item.setText(0,str(name));
      item.setText(1,str(server.statestr).lower() or '');
      cwd = server.cwd;
      if cwd:
        home = os.path.expanduser('~');
        prefix = os.path.commonprefix((home,cwd));
        if prefix:
          cwd = os.path.join('~',cwd[len(prefix)+1:],'');
      item.setText(2,cwd or '');
      script = staterec and staterec.get('script_name');
      script = script and os.path.basename(script);
      item.setText(3,script or '');



#!/usr/bin/env python3

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
#

# A plugin to plot matplotlib / pylab scripts generated by JEN pynodes.
# Gratefully adapted from the matplotlib examples: embedding_in_qt.py
# and embedding_in_qt4.py

# modules that are imported
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from Timba.dmi import *
from Timba import utils
from Timba.Meq import meqds
from Timba.Meq.meqds import mqs
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI import widgets
from MeqGUI.GUI.browsers import *
from MeqGUI import Grid

from PyQt4 import Qt

from .ResultsRange_qt4 import *
from .BufferSizeDialog_qt4 import *

from numpy import arange, sin, cos, pi
import os, sys

# test if matplotlib / pylab is installed
global has_pylab
has_pylab = False
try:
  import matplotlib
  from matplotlib.backends.backend_qt4agg import FigureCanvasQTAgg as FigureCanvas
  from matplotlib.backends.backend_qt4 import NavigationToolbar2QT as NavigationToolbar
  from matplotlib.figure import Figure
  import pylab
  has_pylab = True
except:
  print(' ')
  print('*** matplotlib/pylab not imported! ***')
  print('The system will assume that matplotlib is not present.')

from Timba.utils import verbosity
_dbg = verbosity(0,name='pylab_plotter');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

if has_pylab:
 class MyPylabPlotter(FigureCanvas):
   
   def __init__(self, parent=None, dpi=100):
     self.fig = Figure(dpi=dpi)
     FigureCanvas.__init__(self, self.fig)
#    self.reparent(parent, QPoint(0, 0))
#    self.setParent(Qt.QWidget())
     FigureCanvas.setSizePolicy(self,
                                Qt.QSizePolicy.Expanding,
                                Qt.QSizePolicy.Expanding)
     FigureCanvas.updateGeometry(self)

   def sizeHint(self):
     w, h = self.get_width_height()
     return Qt.QSize(w, h)

   def minimumSizeHint(self):
     return Qt.QSize(10, 10)

   # The following method is adapted from 
   # matplotlib/examples/pythonic_matplotlib.py
   # It shows how you can approximate pylab calls
   # in a 'pythonic' manner which allows the plot
   # to be embedded in the MyPylabPlotter figure.
   def demo_pythonic_matplotlib(self):
     """Simple demo canvas with some sine plots."""
     t = arange(0.0, 1.0, 0.01)
     ax1 = self.fig.add_subplot(211)
     ax1.plot(t, sin(2*pi*t))
     ax1.grid(True)
     ax1.set_ylim( (-2,2) )
     ax1.set_ylabel('1 Hz')
     ax1.set_title('A sine wave or two')

     for label in ax1.get_xticklabels():
         label.set_color('r')

     ax2 = self.fig.add_subplot(212)
     ax2.plot(t, sin(2*2*pi*t))
     ax2.grid(True)
     ax2.set_ylim( (-2,2) )
     ax2.set_ylabel('2 Hz')
     l = ax2.set_xlabel('Hi mom')
     l.set_color('g')
     l.set_fontsize('large')

   # The following method causes a separate pylab figure to  
   # appear. 
   def demo_pylab_figure(self):
     """Simple demo canvas with a sine plot."""
     # create a separate pylab figure
     pylab.figure(1)   
     pylab.subplot(111)
     t = arange(0.0, 3.0, 0.01)
     s = sin(2*pi*t)
     pylab.plot(t, s)
     pylab.grid()
     pylab.title('Demonstration of Matplotlib pylab-style Plugin')
     pylab.xlabel('time')
     pylab.ylabel('sine')
     pylab.show()

   #-------------------------------------------------------------------------

   def make_plot(self, plot_defs):
     """Make a pylab plot from all items in plot_defs.
     """
     from Timba.Contrib.JEN.pylab import PyNodePlot
     PyNodePlot.make_pylab_figure(plot_defs, figob=self.fig)
     return None

#========================================================================

class PylabPlotter(GriddedPlugin):
  """ a class to visualize data from external pylab graphics files """

  _icon = pixmaps.bars3d
  viewer_name = "Pylab Plotter";
  def is_viewable (data):
    return len(data) > 0;
  is_viewable = staticmethod(is_viewable);

  def __init__(self,gw,dataitem,cellspec={},**opts):
    GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    """ a plugin for showing pylab plots """

    self._rec = None;
    self._wtop = None;
    self.dataitem = dataitem
    self.png_number = 0
    self.data_list = []
    self.data_list_labels = []
    self.data_list_length = 0
    self.max_list_length = 50
    self.layout_created = False
    self.counter = -1

    self.reset_plot_stuff()

# back to 'real' work
    if dataitem and dataitem.data is not None:
      self.set_data(dataitem);

  def reset_plot_stuff(self):
    """ resets widgets to None. Needed if we have been putting
        out a message about Cache not containing results, etc
    """
    self._pylab_plotter = None
    self._toolbar = None
    self.results_selector = None
    self.status_label = None
    self.layout_parent = None
    self.layout = None

  def wtop (self):
    """ function needed by Oleg for reasons known only to him! """
    return self._wtop;

  def create_layout_stuff(self):
    """ create grid layouts into which plotter widgets are inserted """
    if self.layout_parent is None or not self.layout_created:
      self.layout_parent = Qt.QWidget(self.wparent())
      self.layout = Qt.QGridLayout(self.layout_parent)
      self.set_widgets(self.layout_parent,self.dataitem.caption,icon=self.icon())
      self.layout_created = True
    self._wtop = self.layout_parent;       

  def set_data (self,dataitem,default_open=None,**opts):
    """ this callback receives data from the meqbrowser, when the
        user has requested a plot. It decides whether the data is
        from a VellSet or visu data record, and  after any
        necessary preprocssing forwards the data to one of
        the functions which does the actual plotting """

    _dprint(3, '** in pylab_plotter:set_data callback')

    if not has_pylab:
      Message = "Matplotlib does not appear to be installed so no plot can be made."
      cache_message = Qt.QLabel(Message,self.wparent())
#     cache_message.setTextFormat(Qt.RichText)
      self._wtop = cache_message
      self.set_widgets(cache_message)
      self.reset_plot_stuff()
      return

    self._rec = dataitem.data;
    _dprint(3, 'set_data: initial self._rec ', self._rec)
# if we are single stepping through requests, Oleg may reset the
# cache, so check for a non-data record situation
    if self._rec is None:
      return
    if isinstance(self._rec, bool):
      return

    self.label = '';  # extra label, filled in if possible
# there's a problem here somewhere ...
    if dmi_typename(self._rec) != 'MeqResult': # data is not already a result?
      # try to put request ID in label
      rq_id_found = False
      data_failure = False
      try:
        if "request_id" in self._rec.cache:
          self.label = "rq " + str(self._rec.cache.request_id);
          rq_id_found = True
        if "result" in self._rec.cache:
          self._rec = self._rec.cache.result; # look for cache.result record
          if not rq_id_found and "request_id" in self._rec:
            self.label = "rq " + str(self._rec.request_id);
        else:
          data_failure = True
        _dprint(3, 'we have req id ', self.label)
      except:
        data_failure = True
      if data_failure:
        _dprint(3, ' we have a data failure')
# cached_result not found, display an empty viewer with a "no result
# in this node record" message (the user can then use the Display with
# menu to switch to a different viewer)
        Message = "No cache result record was found for this node, so no plot can be made."
        cache_message = Qt.QLabel(Message,self.wparent())
#       cache_message.setTextFormat(Qt.RichText)
        self._wtop = cache_message
        self.set_widgets(cache_message)
        self.reset_plot_stuff()

        return
    
# update display with current data
    process_result = self.process_data()

# add this data set to internal list for later replay
    if process_result:
      if self.max_list_length > 0:
        self.data_list.append(self._rec)
        self.data_list_labels.append(self.label)
        if len(self.data_list_labels) > self.max_list_length:
          del self.data_list_labels[0]
          del self.data_list[0]
        if len(self.data_list) != self.data_list_length:
          self.data_list_length = len(self.data_list)
        if self.data_list_length > 1:
          _dprint(3, 'calling adjust_selector')
          self.adjust_selector()

  def process_data (self):
    """ process the actual record structure associated with a Cache result """
    process_result = False
# are we dealing with an pylab result?
    if "plotdefs" in self._rec:
      self.create_layout_stuff()
      self.show_pylab_plot()
      process_result = True

# enable & highlight the cell
    self.enable();
    self.flash_refresh();
    _dprint(3, 'exiting process_data')
    return process_result

  def replay_data (self, data_index):
    """ call to redisplay contents of a result record stored in 
        a results history buffer
    """
    if data_index < len(self.data_list):
      self._rec = self.data_list[data_index]
      self.label = self.data_list_labels[data_index]
      self.results_selector.setLabel(self.label)
      process_result = self.process_data()

  def show_pylab_plot(self, store_rec=True):
    """ process incoming data and attributes into the
        appropriate type of plot """

    # if we are single stepping through requests, Oleg may reset the
    # cache, so check for a non-data record situation
    if store_rec and isinstance(self._rec, bool):
      return

    pylab_record = self._rec.plotdefs
    if not self._pylab_plotter is None:
#     self._pylab_plotter.reparent(Qt.QWidget(), 0, Qt.QPoint())
      self._pylab_plotter.setParent(Qt.QWidget())
      self._pylab_plotter  = None
    if not self._toolbar is None:
#     self._toolbar.reparent(Qt.QWidget(), 0, Qt.QPoint())
      self._toolbar.setParent(Qt.QWidget())
      self._toolbar = None
    if self._pylab_plotter is None:
      self._pylab_plotter = MyPylabPlotter(parent=self.layout_parent)
      self.layout.addWidget(self._pylab_plotter,1,0)
      self._pylab_plotter.show()
    if self._toolbar is None:
      self._toolbar = NavigationToolbar(self._pylab_plotter, self.layout_parent)
      self._toolbar.setSizePolicy(Qt.QSizePolicy.Expanding,Qt.QSizePolicy.Fixed)
      self.layout.addWidget(self._toolbar,0,0)
      self._toolbar.show()
    self._pylab_plotter.make_plot(pylab_record)
    
  # end show_pylab_plot()


  def set_results_buffer (self, result_value):
    """ callback to set the number of results records that can be
        stored in a results history buffer 
    """ 
    if result_value < 0:
      return
    self.max_list_length = result_value
    if len(self.data_list_labels) > self.max_list_length:
      differ = len(self.data_list_labels) - self.max_list_length
      for i in range(differ):
        del self.data_list_labels[0]
        del self.data_list[0]

    if len(self.data_list) != self.data_list_length:
      self.data_list_length = len(self.data_list)

    self.show_pylab_plot(store_rec=False)

  def adjust_selector (self):
    """ instantiate and/or adjust contents of ResultsRange object """
    if self.results_selector is None:
      self.results_selector = ResultsRange(self.layout_parent)
      self.results_selector.setMaxValue(self.max_list_length)
      self.results_selector.set_offset_index(0)
      self.results_selector.setSizePolicy(Qt.QSizePolicy.Expanding,Qt.QSizePolicy.Minimum)
      self.layout.addWidget(self.results_selector,2,0)
      self.results_selector.show()
      QObject.connect(self.results_selector, PYSIGNAL('result_index'), self.replay_data)
      QObject.connect(self.results_selector, PYSIGNAL('adjust_results_buffer_size'), self.set_results_buffer)
    self.results_selector.set_emit(False)
    self.results_selector.setRange(self.data_list_length-1)
    self.results_selector.setLabel(self.label)
    self.results_selector.set_emit(True)

#Grid.Services.registerViewer(dmi_type('MeqResult',record),PylabPlotter,priority=10)
Grid.Services.registerViewer(meqds.NodeClass(),PylabPlotter,priority=50)

########################################
# stuff for testing
########################################
if has_pylab:
  class ApplicationWindow(QMainWindow):
    def __init__(self):
        Qt.QMainWindow.__init__(self, None)
#                            "application main window",
#                            Qt.WType_TopLevel | Qt.WDestructiveClose)

        self.file_menu = Qt.QMenu('&File',self)
        self.file_menu.addAction('&Quit', self.fileQuit, Qt.Qt.CTRL + Qt.Qt.Key_Q)
        self.menuBar().addMenu(self.file_menu)

        self.help_menu = Qt.QMenu('&Help', self)
        self.menuBar().addSeparator()
        self.menuBar().addMenu(self.help_menu)

        self.help_menu.addAction('&About', self.about)

        self.main_widget = Qt.QWidget(self)

        l = Qt.QVBoxLayout(self.main_widget)
        sc = MyPylabPlotter(self.main_widget, dpi=100)
        toolbar = NavigationToolbar(sc, self.main_widget)
        toolbar.setSizePolicy(Qt.QSizePolicy.Expanding,Qt.QSizePolicy.Fixed)
        l.addWidget(sc)
        l.addWidget(toolbar)

        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)

# Produce a plot that is contained in MyPylabPlotter
        sc.demo_pythonic_matplotlib()
# Produce a plot that is separate from MyPylabPlotter
        sc.demo_pylab_figure()

    def fileQuit(self):
        qApp.exit(0)

    def closeEvent(self, ce):
        self.fileQuit()

    def about(self):
        progname = os.path.basename(sys.argv[0])
        progversion = "0.1"
        QMessageBox.about(self, "About %s" % progname,
"""%(prog)s version %(version)s
Copyright \N{COPYRIGHT SIGN} 2005 Florent Rougon

This program is a simple example of a Qt application embedding matplotlib
canvases.

It may be used and modified with no restriction; raw copies as well as
modified versions may be distributed without limitation."""
                          % {"prog": progname, "version": progversion})


def main( argv ):
  if has_pylab:
    app = Qt.QApplication(sys.argv)
    aw = ApplicationWindow()
#   app.setMainWidget(aw)
    aw.show()
    sys.exit(app.exec_())
  else:
    print(' ')
    print('**** Sorry! It looks like matplotlib/pylab is not available! ****')


# Admire
if __name__ == '__main__':
  main(sys.argv)





#!/usr/bin/env python3

# modules that are imported

#% $Id: result_plotter.py 6419 2008-10-03 10:17:35Z oms $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from math import sin
from math import cos
from math import pow
from math import sqrt

from qwt.qt.QtGui import QApplication,QHBoxLayout, QLabel, QSizePolicy, QSpacerItem
from qwt.qt.QtGui import QWidget
from qwt.qt.QtCore import Qt
from qwt import QwtSymbol, QwtPlotCurve

HAS_TIMBA = False
# modules that are imported
try:
  from Timba.dmi import *
  from Timba import utils
  from Timba.Meq import meqds
  from Timba.Meq.meqds import mqs
  from MeqGUI.GUI.pixmaps import pixmaps
  from MeqGUI.GUI import widgets
  from MeqGUI.GUI.browsers import *
  from MeqGUI import Grid
  
  from MeqGUI.Plugins.display_image_qt5 import QwtImageDisplay
  from MeqGUI.Plugins.QwtPlotImage_qt5 import QwtPlotImage
  from MeqGUI.Plugins.QwtColorBar_qt5 import QwtColorBar
  from MeqGUI.Plugins.SpectrumData import SpectrumData
  from MeqGUI.Plugins.VellsData import VellsData
  from MeqGUI.Plugins.SolverData import SolverData
  from MeqGUI.Plugins.ND_Controller_qt5 import ND_Controller
  from MeqGUI.Plugins.ResultsRange_qt5 import ResultsRange
  import MeqGUI.Plugins.plotting_functions_qt5 as plot_func
  from Timba.utils import verbosity
  _dbg = verbosity(0,name='result_plotter');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
  import traceback
  traceback.print_exc()
  print("Cannot import gui and plotting utilities. Plotting will not be available")

global has_vtk
has_vtk = False


class ResultPlotter(GriddedPlugin):
  """ a class to visualize data, VellSets or visu data, that is 
      contained within a node's cache_result record. Objects of 
      this class are launched from the meqbrowser GUI """

  _icon = pixmaps.bars3d
  viewer_name = "Result Plotter";
  def is_viewable (data):
    return len(data) > 0;
  is_viewable = staticmethod(is_viewable);

# the following global tables replicate similar tables found in the
# realvsimag plotter. The idea is that the system first checks the
# contents of 'visu' plot records against these tables here 
# during tree traversal. Otherwise every leaf node would issue
# warnings about the same unacceptable parameters - which would
# really irritate the user. The 'check_attributes' function defined
# below does the work.
  color_table = {
        'none': None,
        'black': Qt.black,
        'blue': Qt.blue,
        'cyan': Qt.cyan,
        'gray': Qt.gray,
        'green': Qt.green,
        'magenta': Qt.magenta,
        'red': Qt.red,
        'white': Qt.white,
        'yellow': Qt.yellow,
        'darkBlue' : Qt.darkBlue,
        'darkCyan' : Qt.darkCyan,
        'darkGray' : Qt.darkGray,
        'darkGreen' : Qt.darkGreen,
        'darkMagenta' : Qt.darkMagenta,
        'darkRed' : Qt.darkRed,
        'darkYellow' : Qt.darkYellow,
        'lightGray' : Qt.lightGray,
        }

  symbol_table = {
#       'none': QwtSymbol.None,
        'rectangle': QwtSymbol.Rect,
        'square': QwtSymbol.Rect,
        'ellipse': QwtSymbol.Ellipse,
        'dot': QwtSymbol.Ellipse,
        'circle': QwtSymbol.Ellipse,
        'xcross': QwtSymbol.XCross,
        'cross': QwtSymbol.Cross,
        'triangle': QwtSymbol.Triangle,
        'diamond': QwtSymbol.Diamond,
        }

  line_style_table = {
        'none': QwtPlotCurve.NoCurve,
        'lines' : QwtPlotCurve.Lines,
        'dots' : QwtPlotCurve.Dots,
        'SolidLine' : Qt.SolidLine,
        'DashLine' : Qt.DashLine,
        'DotLine' : Qt.DotLine,
        'DashDotLine' : Qt.DashDotLine,
        'DashDotDotLine' : Qt.DashDotDotLine,
        'solidline' : Qt.SolidLine,
        'dashline' : Qt.DashLine,
        'dotline' : Qt.DotLine,
        'dashdotline' : Qt.DashDotLine,
        'dashdotdotline' : Qt.DashDotDotLine,
        }
  
  def __init__(self,gw,dataitem,cellspec={},**opts):
    GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    """ Instantiate HippoDraw objects that are used to control
        various aspects of plotting, if the hippo plotter
        is instantiated.
    """
    self._rec = None;
    self._plot_type = None
    self._wtop = None;
    self.dataitem = dataitem
    self._attributes_checked = False
    self._vells_data = None
    self._solver_data = None
    self.num_possible_ND_axes = None
    self.active_image_index = None
    self._spectrum_data = None
    self.png_number = 0
    self.data_list = []
    self.data_list_labels = []
    self.data_list_length = 0
    self.max_list_length = 50
    self._window_controller = None
    self.array_shape = None
    self.actual_rank = None
    self.plot_3D = False
    self.ND_plotter = None
    self.ND_labels = None
    self.ND_parms = None 
    self.status_label = None
    self.layout_created = False

    self.reset_plot_stuff()

# back to 'real' work
    if dataitem and dataitem.data is not None:
      self.set_data(dataitem);

  def reset_plot_stuff(self):
    """ resets widgets to None. Needed if we have been putting
        out a message about Cache not containing results, etc
    """
    self._visu_plotter = None
    self.QTextEdit = None
    self.colorbar = {}
    self.results_selector = None
    self.spectrum_node_selector = None
    self.status_label = None
    self.ND_Controls = None
    self.ND_plotter = None
    self.layout_parent = None
    self.layout = None


  def __del__(self):
    if self._window_controller:
      self._window_controller.closeAllWindows()
                                                                                           
  def wtop (self):
    """ function needed by Oleg for reasons known only to him! """
    return self._wtop;

  def plotSpectra(self, leaf_record):
    """ stores and plots data for a visu Spectra node """
    self._spectrum_data = None
    if self._spectrum_data is None:
      (self._data_labels, self._string_tag) = self._visu_plotter.getSpectrumTags()
      self._spectrum_data = SpectrumData(self._data_labels, self._string_tag)
    if 'value' in leaf_record:
      self._data_values = leaf_record['value']

# store the data
    self._spectrum_data.StoreSpectrumData(self._data_values)

# test and update the context menu
    plot_menus = self._spectrum_data.getMenuLabels()
    self._visu_plotter.setSpectrumMenuItems(plot_menus)
    spectrum_menu_items = len(plot_menus)
    if spectrum_menu_items > 2: 
      marker_parms = self._spectrum_data.getMarkerParms()
      marker_labels = self._spectrum_data.getMarkerLabels()
      self._visu_plotter.setSpectrumMarkers(marker_parms, marker_labels)

# plot active instance of array
    if self.active_image_index is None or self.active_image_index > spectrum_menu_items - 1:
      self.active_image_index = spectrum_menu_items - 1
      self._spectrum_data.setActivePlot(self.active_image_index)
    plot_label = self._spectrum_data.getPlotLabel()
    plot_data = self._spectrum_data.getActivePlotArray()
    self._visu_plotter.array_plot(plot_data, data_label=plot_label,flip_axes=False)


  def check_attributes(self, attributes):
     """ check parameters of plot attributes against allowable values """
     plot_parms = None
     if 'plot' in attributes:
       plot_parms = attributes.get('plot')
       if 'attrib' in plot_parms:
         temp_parms = plot_parms.get('attrib')
         plot_parms = temp_parms
       if 'color' in plot_parms:
         plot_color = plot_parms.get('color')
         if plot_color not in self.color_table:
           Message = str(plot_color) + " is not a valid color.\n Using blue by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['color'] = "blue"
       if 'mean_circle_color' in plot_parms:
         plot_color = plot_parms.get('mean_circle_color')
         if plot_color not in self.color_table:
           Message = str(plot_color) + " is not a valid color.\n Using blue by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['mean_circle_color'] = "blue"
       if 'stddev_circle_color' in plot_parms:
         plot_color = plot_parms.get('stddev_circle_color')
         if plot_color not in self.color_table:
           Message = str(plot_color) + " is not a valid color.\n Using blue by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['stddev_circle_color'] = "blue"
       if 'line_style' in plot_parms:
         plot_line_style = plot_parms.get('line_style')
         if plot_line_style not in self.line_style_table:
           Message = str(plot_line_style) + " is not a valid line style.\n Using dots by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['line_style'] = "dots"
       if 'mean_circle_style' in plot_parms:
         plot_line_style = plot_parms.get('mean_circle_style')
         if plot_line_style not in self.line_style_table:
           Message = str(plot_line_style) + " is not a valid line style for mean circles.\n Using lines by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['mean_circle_style'] = "lines"
       if 'stddev_circle_style' in plot_parms:
         plot_line_style = plot_parms.get('stddev_circle_style')
         if plot_line_style not in self.line_style_table:
           Message = str(plot_line_style) + " is not a valid line style for stddev circles.\n Using DotLine by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['stddev_circle_style'] = "DotLine"
       if 'symbol' in plot_parms:
         plot_symbol = plot_parms.get('symbol')
         if plot_symbol not in self.symbol_table:
           Message = str(plot_symbol) + " is not a valid symbol.\n Using circle by default"
           mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
           plot_parms['symbol'] = "circle"
  # check_attributes

#
# tree traversal code adapted from the pasteur institute python 
# programming course chapter on recursive data structures at
# http://www.pasteur.fr/formation/infobio/python/ch13s03.html
#

  def do_prework(self, node, attribute_list):
    """ do any processing before actual handling of data in a leaf node """
# we check if a plotter has been constructed - 
    if isinstance(node, dict) and self._visu_plotter is None:
      if len(attribute_list) == 0 and 'attrib' in node:
        if len(node['attrib']) > 0:
          attrib_parms = node['attrib']
          plot_parms = attrib_parms.get('plot')
          if 'plot_type' in plot_parms:
            self._plot_type = plot_parms.get('plot_type')
          if 'type' in plot_parms:
            self._plot_type = plot_parms.get('type')
          if 'results_buffer' in plot_parms:
            self.max_list_length = plot_parms.get('results_buffer')
      else:
# first get plot_type at first possible point in list - nearest root
        list_length = len(attribute_list)
        for i in range(list_length):
          attrib_parms = attribute_list[i]
 
          if 'plot' in attrib_parms:
            plot_parms = attrib_parms.get('plot')
            if 'attrib' in plot_parms:
              temp_parms = plot_parms.get('attrib')
              plot_parms = temp_parms
            if 'results_buffer' in plot_parms:
              self.max_list_length = plot_parms.get('results_buffer')
            if 'plot_type' in plot_parms:
              self._plot_type = plot_parms.get('plot_type')
              break
            if 'type' in plot_parms:
              self._plot_type = plot_parms.get('type')
              break

# create grid layout for widgets
      self.create_layout_stuff()

      if self._plot_type == 'spectra':
        self.create_2D_plotter()

      if self._plot_type == 'realvsimag':

        self._visu_plotter = realvsimag_plotter(self._plot_type,parent=self.layout_parent)
        self._visu_plotter.plot.save_display.connect(self.grab_display)
        self.layout.addWidget(self._visu_plotter.plot, 0, 1)
        self._visu_plotter.plot.show()


  def do_postwork(self, node):
    """ do any processing needed after data in a leaf node has been handled """


  def is_leaf(self, node):
    """ tests if a node is actually a leaf node """
    if 'value' in node:
      candidate_leaf = node['value']
      if isinstance(candidate_leaf, list):
# check if list contents are a dict
        for i in range(len(candidate_leaf)):
           if isinstance(candidate_leaf[i], dict):
             return False
        return True
    else:
      return False

  def do_leafwork(self, leaf, attrib_list):
    """ method which does actual plotting at a leaf node """

# If we arrive here without having gotten a plot type
# it is because the user specified an invalid type somehow.
# Post a message and select the default. 
    if self._visu_plotter is None:
      message = None
      if not self._plot_type is None:
        Message = self._plot_type + " is not a valid plot type.\n Using realvsimag by dafault." 
      else:
        Message = "Failure to find a valid plot type.\n Using realvsimag by default."
      mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)
      self._plot_type = "realvsimag"
      self._visu_plotter = realvsimag_plotter(self._plot_type,parent=self.wparent())
      self._visu_plotter.plot.save_display.connect(self.grab_display)
      self.set_widgets(self._visu_plotter,self.dataitem.caption,icon=self.icon())
      self._wtop = self._visu_plotter.plot;  # plot widget is our top widget

# now do the plotting
    if self._plot_type == 'spectra':
      if self.first_leaf_node:
        self.leaf_node_list = []
        self.list_attrib_lists = []
        self.list_labels = []
        self._visu_plotter.plot_data(leaf, attrib_list, label=self.label)
        self.plotSpectra(leaf)
        self.leaf_node_list.append(leaf)
        self.list_attrib_lists.append(attrib_list)
        self.list_labels.append(self.label)
        self.first_leaf_node = False
      else:
# I'm not sure when/how one ends up in the 'else' section
        self.leaf_node_list.append(leaf)
        self.list_attrib_lists.append(attrib_list)
        self.list_labels.append(self.label)
        self.adjust_spectrum_selector()
    else:
      self._visu_plotter.plot_data(leaf, attrib_list, label=self.label)

  # do_leafwork

  def tree_traversal (self, node, label=None, attribute_list=None):
    """ routine to do a recursive tree traversal of a Visu plot tree """
    is_root = False
    if label is None:
      label = 'root'
      is_root = True
      node['plot_label'] = ''
      
    if attribute_list is None:
      attribute_list = []
    if isinstance(node, dict):
      if self._visu_plotter is None and not is_root:
# call the do_prework method to do any actions needed before
# an actual leaf node performs plotting operations
        self.do_prework(node, attribute_list)
# test if this node is a leaf
      if not self.is_leaf(node):
        if 'label' in node:
          if not node['label'] == label:
            if isinstance(node['label'], tuple):
              temp = list(node['label'])
              for j in range(0, len(temp)):
                tmp = label + '\n' + temp[j] 
                temp[j] = tmp
              node['plot_label'] = tuple(temp)
            else:
              temp = label + '\n' + node['label']
              node['plot_label'] = temp
        if 'attrib' in node and len(node['attrib']) > 0:
          if not self._attributes_checked:
            self.check_attributes(node['attrib'])
          attribute_list.append(node['attrib'])
        else:
          if is_root:
            attrib = {}
            plot_spec = {}
            plot_spec['plot_type'] = 'realvsimag'
            plot_spec['mean_circle'] = False
            plot_spec['mean_arrow'] = False
            plot_spec['stddev_circle'] = False
            attrib['plot'] = plot_spec
            attribute_list.append(attrib)
# if not a leaf, and we find a 'value' field, then call
# recursive method 'tree_traversal'
        if 'value' in node:
          self.tree_traversal(node['value'], node['plot_label'], attribute_list)
      else:
        try:
          print('tree: leaf node has label(s) ', node['plot_label'])
          print('tree: leaf node has incoming label ', label)
        except:
          print('node label field expected, not found, so am exiting')
          Message = "Failure of result_plotter tree-traversal.\n Result_plotter does not yet work with MeqHistoryCollect nodes."
          mb_reporter = QMessageBox.warning(None, "ResultPlotter", Message)
          return

        if is_root and 'attrib' in node and len(node['attrib']) > 0:
          if not self._attributes_checked:
            self.check_attributes(node['attrib'])
          attribute_list.append(node['attrib'])

# call the do_prework method to do any actions needed before
# actual leaf node performs plotting operations
          self.do_prework(node, attribute_list)

# if all tests are passed and this is a leaf, than do actual plotting work
        self.do_leafwork(node,attribute_list)

# no post work at the present time
#      self.do_postwork(node)

# if we are at a level where we encounter a bunch of nodes in a list
# then we must preform a recursive tree traversal starting with
# each element in the list
    if isinstance(node, list):
      for i in range(len(node)):
        temp_label = None
        temp_list = attribute_list[:] 
        if isinstance(label, tuple):
          temp_label = label[i]
        else:
          temp_label = label
          
        if isinstance(node[i], dict):
          if 'label' in node[i]:
            if isinstance(node[i]['label'], tuple):
              temp = list(node[i]['label'])
              for j in range(0, len(temp)):
                tmp = temp_label + '\n' + temp[j]
                temp[j] = tmp
              node[i]['plot_label'] = tuple(temp)
            else:
              temp = label + '\n' + node[i]['label']
              node[i]['plot_label'] = temp
          if 'attrib' in node[i]:
            if len(node[i]['attrib']) > 0:
              if not self._attributes_checked:
                self.check_attributes(node[i]['attrib'])
              temp_list.append(node[i]['attrib'])
          self.tree_traversal(node[i], node[i]['label'], temp_list)

  def display_visu_data (self):
    """ extract group_label key from incoming visu data record and
        create a visu_plotter object to plot the data 
    """
# traverse the plot record tree and retrieve data
    self.first_leaf_node = True
    self.tree_traversal( self._rec.visu)
# now update the plot for 'realvsimag', 'errors' or 'standalone' plot
    if not self._visu_plotter is None and not self._plot_type == 'spectra':
      self._visu_plotter.update_plot()
      self._visu_plotter.reset_data_collectors()

  def create_layout_stuff(self):
    """ create grid layouts into which plotter widgets are inserted """
    if self.layout_parent is None or not self.layout_created:
      self.layout_parent = QWidget(self.wparent())
      self.layout = QGridLayout(self.layout_parent)
      self.QTextEdit = QTextEdit(self.layout_parent)
      self.layout.addWidget(self.QTextEdit, 0, 1)
      self.QTextEdit.hide()
      self.QTextEdit.setReadOnly(True)
      self.set_widgets(self.layout_parent,self.dataitem.caption,icon=self.icon())
      self.layout_created = True
    self._wtop = self.layout_parent;       

  def create_2D_plotter(self):
    if not self.ND_plotter is None:
      self.ND_plotter.close()
# the following statement causes a crash
#     self.ND_plotter.setParent(Qt.QWidget())
      self.ND_plotter = None

    self.plotPrinter = None
    self._visu_plotter = plot_func.create_2D_Plotters(self.layout, self.layout_parent)
    self._visu_plotter.handle_menu_id.connect(self.update_vells_display)
    self._visu_plotter.handle_spectrum_menu_id.connect(self.update_spectrum_display)
    self._visu_plotter.colorbar_needed.connect(self.set_ColorBar)
    self._visu_plotter.show_ND_Controller.connect(self.ND_controller_showDisplay)
    self._visu_plotter.show_3D_Display.connect(self.show_3D_Display)
#   self._visu_plotter.do_print.connect(self.plotPrinter.do_print)
    self._visu_plotter.save_display.connect(self.grab_display)
    self._visu_plotter.full_vells_image.connect(self.request_full_image)
  # create_2D_plotter

  def grab_display(self, title):
    self.png_number = self.png_number + 1
    png_str = str(self.png_number)
    if title is None:
      save_file = self._node_name + './meqbrowser' + png_str + '.png'
    else:
      save_file = self._node_name + title + png_str + '.png'
    save_file_no_space= save_file.replace(' ','_')
    result = Qt.QPixmap.grabWidget(self.layout_parent).save(save_file_no_space, "PNG")

    
  def set_data (self,dataitem,default_open=None,**opts):
    """ this callback receives data from the meqbrowser, when the
        user has requested a plot. It decides whether the data is
        from a VellSet or visu data record, and  after any
        necessary preprocssing forwards the data to one of
        the functions which does the actual plotting """

    self._rec = dataitem.data;
# if we are single stepping through requests, Oleg may reset the
# cache, so check for a non-data record situation
    if self._rec is None:
      return
    if isinstance(self._rec, bool):
      return

# if this node is a Composer, it might have a plot_label which we
# want to use later
    try:
      self._plot_label = self._rec.plot_label
    except:
      self._plot_label = None

# get the name of the node - used for saving images
    try:
      self._node_name = self._rec.name + '_'
    except:
      self._node_name = ''

    self.label = '';  # extra label, filled in if possible
# there's a problem here somewhere ...
    if dmi_typename(self._rec) != 'MeqResult': # data is not already a result?
      # try to put request ID in label
      rq_id_found = False
      data_failure = False
      try:
        if "request_id" in self._rec.cache:
          self.label = "rq " + str(self._rec.cache.request_id);
          rq_id_found = True
        if "result" in self._rec.cache:
          self._rec = self._rec.cache.result; # look for cache.result record
          if not rq_id_found and "request_id" in self._rec:
            self.label = "rq " + str(self._rec.request_id);
        else:
          data_failure = True
      except:
        data_failure = True
      if data_failure:
# cached_result not found, display an empty viewer with a "no result
# in this node record" message (the user can then use the Display with
# menu to switch to a different viewer)
        Message = "No cache result record was found for this node, so no plot can be made."
        cache_message = QLabel(Message,self.wparent())
        cache_message.setTextFormat(Qt.RichText)
        self._wtop = cache_message
        self.set_widgets(cache_message)
        self.reset_plot_stuff()

        return

# check for label matching - this is necessary as a node may be called
# by multiple parents, and thus send the same result to the display
# more than once
    label_found = False
    if len(self.data_list_labels) > 0:
      for i in range(len(self.data_list_labels)):
        if self.data_list_labels[i] == self.label:
          label_found = True
    if not label_found and self.max_list_length > 0:
# update display with current data
      process_result = self.process_data()
# add this data set to internal list for later replay if result was OK
      if process_result:
        self.data_list.append(self._rec)
        self.data_list_labels.append(self.label)
        if len(self.data_list_labels) > self.max_list_length:
          del self.data_list_labels[0]
          del self.data_list[0]
        if len(self.data_list) != self.data_list_length:
          self.data_list_length = len(self.data_list)
        if self.data_list_length > 1:
          self.adjust_selector()

  def process_data (self):
    """ process the actual record structure associated with a Cache result """
    process_result = False
# are we dealing with Vellsets?
    if "dims" in self._rec:
      pass
#     print('*** dims field exists ', self._rec.dims)
    if "vellsets" in self._rec or "solver_result" in self._rec:
      self.create_layout_stuff()
      if "vellsets" in self._rec:
        process_result = self.plot_vells_data()
        if not process_result:
          Message = "The result record for this node had no valid data, so no plot can be made."
          cache_message = QLabel(Message,self.wparent())
          cache_message.setTextFormat(Qt.RichText)
          self._wtop = cache_message
          self.set_widgets(cache_message)
          self.reset_plot_stuff()
      else:
        if self._visu_plotter is None:
          self.create_2D_plotter()
        self.plot_solver()
        process_result = True

# otherwise we are dealing with a set of visualization data
    else:
      if "visu" in self._rec:
# do plotting of visualization data
        self.display_visu_data()
        process_result = True

# enable & highlight the cell
    self.enable();
    self.flash_refresh();
    return process_result

  def replay_data (self, data_index):
    """ callback to redisplay contents of a result record stored in 
        a results history buffer
    """
    if data_index < len(self.data_list):
      self._rec = self.data_list[data_index]
      self.label = self.data_list_labels[data_index]
      self.results_selector.setLabel(self.label)
      process_result = self.process_data()

  def select_spectrum_node (self, data_index):
    """ callback to redisplay contents of a spectrum leaf node stored in 
        a list of leaf nodes
    """
    if data_index < len(self.leaf_node_list):
      leaf = self.leaf_node_list[data_index]
      attrib_list = self.list_attrib_lists[data_index]
      label = self.list_labels[data_index]
      self._visu_plotter.plot_data(leaf, attrib_list,label)
      self.plotSpectra(leaf)

  def plot_vells_data (self, store_rec=True):
      """ process incoming vells data and attributes into the
          appropriate type of plot """

# if we are single stepping through requests, Oleg may reset the
# cache, so check for a non-data record situation
      if store_rec and isinstance(self._rec, bool):
        return


# store the data
      if self._vells_data is None:
        self._vells_data = VellsData()

      self._vells_data.set_3D_Display(False)

      if store_rec:
        self._vells_data.setInitialSelection(False)
        if not self._plot_label is None:
          self._vells_data.set_exterior_plot_label(self._plot_label)
        self._vells_data.StoreVellsData(self._rec,self.label)

      if self._vells_data.isVellsScalar():
        if not self._visu_plotter is None:
          self._visu_plotter.hide()
        Message =  self._vells_data.getScalarString()
        if Message.find("has no data") > -1:
          return False
        else:
          self.QTextEdit.setText(Message)
          self.QTextEdit.show()
          return True

      self._vells_plot = True

      if self._visu_plotter is None:
        if not self.ND_Controls is None:
          self.ND_Controls.setParent(Qt.QWidget())
          self.ND_Controls = None
        ranks = self._vells_data.getActiveDataRanks()
        self.create_2D_plotter()
        self._vells_data.setInitialSelectedAxes(ranks[1],ranks[2], reset=True)

      self._visu_plotter.setVellsPlot(self._vells_plot)

# vells axes can change even if data shape has not changed
# so we need to always call this function
      self.update_display_control()

# generate basic menu
      self._visu_plotter.initVellsContextMenu()

# do we have flags for data?	  
      self.test_for_flags()

# test and update the context menu
      menu_data = self._vells_data.getMenuData()
      menu_labels = menu_data[0]
      vells_menu_items = len(menu_labels)
      if vells_menu_items > 1:
        self._visu_plotter.setMenuItems(menu_data)

      if not menu_data[4] is None and len(menu_data[4]) > 0:
        self._visu_plotter.setBigArrays(menu_data[4])

# reset color bar request - just in case
      self._visu_plotter.reset_color_bar(True)

# plot the appropriate plane / perturbed value
# we make a deep copy as the plot array may be modified by flagging etc
# during plotting
      plot_data = self._vells_data.getActiveData().copy()

      # get initial axis parameters
      axis_parms =  self._vells_data.getActiveAxisParms()
      self._visu_plotter.setAxisParms(axis_parms)
      plot_label = self._vells_data.getPlotLabel()
      if not self.test_vells_scalar(plot_data, plot_label):
        black_colour = 0
        self._visu_plotter.setFlagColour(black_colour)
        self._visu_plotter.plot_vells_array(plot_data, plot_label)

      return True
    # end plot_vells_data()

  def test_for_flags(self):
    flag_plane = self._vells_data.getActivePlane()
    if self._vells_data.activePlaneHasFlags():
      self._visu_plotter.set_flag_toggles(flag_plane, True)
      self._visu_plotter.setFlagsData(self._vells_data.getActiveFlagData())
    else:
      self._visu_plotter.set_flag_toggles(flag_plane, False)
      self._visu_plotter.unsetFlagsData()

  def request_full_image(self,signal):
    """ request a full filled-in image from the Vells """
    self._vells_data.request_full_image(signal)
    plot_data = self._vells_data.getActiveData().copy()
    # get initial axis parameters
    axis_parms =  self._vells_data.getActiveAxisParms()
    self._visu_plotter.setAxisParms(axis_parms)
    plot_label = self._vells_data.getPlotLabel()
    if signal:
      if not self.test_vells_scalar(plot_data, plot_label):
        white_colour = 255
        self._visu_plotter.setFlagColour(white_colour)
    else:
      self.test_for_flags()
      black_colour = 0
      self._visu_plotter.setFlagColour(black_colour)
    self._visu_plotter.plot_vells_array(plot_data, plot_label)

  def update_display_control (self):
      vells_data_parms = self._vells_data.getVellsDataParms()
      vells_axis_parms = vells_data_parms[0]
      axis_labels = vells_data_parms[1]
      self._visu_plotter.setVellsParms(vells_axis_parms, axis_labels)
      display_change = False
      if vells_data_parms[2] != self.num_possible_ND_axes:
        self.num_possible_ND_axes = vells_data_parms[2]
        display_change = True

      if vells_data_parms[3] != self.array_shape:
        self.array_shape = vells_data_parms[3]
        display_change = True

      ranks = self._vells_data.getActiveDataRanks()
      actual_rank = ranks[0]
      self.array_rank = ranks[1]
      self.array_shape = ranks[2]
      if self.actual_rank != actual_rank:
        self.actual_rank = actual_rank
        display_change = True
      self._visu_plotter.set_original_array_rank(self.actual_rank)
      if self.actual_rank > 2 and self.ND_Controls is None:
         display_change = True
      if self.actual_rank <= 2 and not self.ND_Controls is None:
         self.ND_Controls.hide()
      if display_change and self.actual_rank > 2:
        # store for later use
        self.ND_labels = axis_labels
        self.ND_parms = vells_axis_parms 
        self.set_ND_controls(labels=axis_labels, parms=vells_axis_parms)
      else:
        if len(vells_axis_parms) <= 2 or self.num_possible_ND_axes <= 2 or self.actual_rank <= 2:
          if not self.ND_Controls is None:
            self.ND_Controls.hide()

  def plot_solver (self):
    """ plots data from a MeqSolver node """
    if self._solver_data is None:
        self._solver_data = SolverData()
# store the data
    self._solver_data.StoreSolverData(self._rec,self.label)

# refresh plotter in case of solver data
    self._visu_plotter.cleanup()

# try calculating condition numbers
    if self._solver_data.processCovarArray():
      self._visu_plotter.set_condition_numbers(self._solver_data.getConditionNumbers())
#   else:
#     Message = "Failure to calculate Covariance Matrix condition number!\nIs the numpy package installed?"
#     mb_reporter = Qt.QMessageBox.warning(None, "ResultPlotter", Message)

    if self._solver_data.calculateCovarEigenVectors():
      self._visu_plotter.set_eigenvectors(self._solver_data.getEigenVectors())

# retrieve solver data
    self._solver_array =  self._solver_data.getSolverData()
    self._visu_plotter.set_solver_metrics(self._solver_data.getSolverMetrics())
    self._visu_plotter.set_solver_labels(self._solver_data.getSolverLabels())
    shape = self._solver_array.shape
    title = ''
    if shape[1] > 1:
      self._x_title = 'Solvable Coeffs'
      self._y_title = 'Iteration Nr'
      title = self.label + " Solver Incremental Solutions"
    else:
      self._y_title = 'Value'
      self._x_title = 'Iteration Nr'
      title = self.label + " Solver Incremental Solution"
    self._visu_plotter.array_plot(self._solver_array,data_label=title)

  def test_vells_scalar (self, data_array, data_label):
    """ test if incoming Vells contains only a scalar value """
# do we have a scalar?
    is_scalar = False
    scalar_data = 0.0
    try:
      shape = data_array.shape
    except:
      is_scalar = True
      scalar_data = data_array
    if not is_scalar and len(shape) == 1:
      if shape[0] == 1:
        is_scalar = True
        scalar_data = data_array[0]
    if not is_scalar:
      test_scalar = True
      for i in range(len(shape)):
        if shape[i] > 1:
          test_scalar = False
      is_scalar = test_scalar
      if is_scalar:
        scalar_data = data_array
    if is_scalar:
      self._visu_plotter.report_scalar_value(data_label, scalar_data)
      return True
    else:
      return False

  def update_vells_display (self, menuid):
    """ callback to handle a request from the lower level 
        display_image.py code for different Vells data """
     
    self._vells_data.unravelMenuId(menuid)
    plot_label = self._vells_data.getPlotLabel()
    plot_data = self._vells_data.getActiveData().copy()

# do we have flags for data?	  
    self.test_for_flags()

    if self._vells_data.getShapeChange():
      self.update_display_control()
      axis_parms =  self._vells_data.getActiveAxisParms()
      self._visu_plotter.setAxisParms(axis_parms)
    self._visu_plotter.reset_color_bar(True)
    if self._vells_data.isVellsScalar():
#     self._visu_plotter.report_scalar_value(plot_label, plot_data)
      if not self._visu_plotter is None:
        self._visu_plotter.hide()
      Message =  self._vells_data.getScalarString()
      if Message.find("has no data") > -1:
        return 
      else:
        self.QTextEdit.setText(Message)
        self.QTextEdit.show()
        return
    else:
      if not self.test_vells_scalar(plot_data, plot_label):
        if not self.QTextEdit is None:
          self.QTextEdit.hide()
        if not self._visu_plotter is None:
          self._visu_plotter.show()
        black_colour = 0
        self._visu_plotter.setFlagColour(black_colour)
        self._visu_plotter.plot_vells_array(plot_data, plot_label)

  def update_spectrum_display(self, menuid):
    """ callback to handle a request from the lower level 
        display_image.py code for different Spectrum data 
    """
    self._spectrum_data.setActivePlot(menuid)
    plot_label = self._spectrum_data.getPlotLabel()
    plot_data = self._spectrum_data.getActivePlotArray()
    self._visu_plotter.array_plot(plot_data, data_label=plot_label, flip_axes=False)

  def setSelectedAxes (self,first_axis, second_axis, third_axis=-1):
    """ callback to handle a request from the N-dimensional
        controller to set new (sub)axes for the Vells display 
    """
    if self._vells_plot:
      self._vells_data.setSelectedAxes(first_axis, second_axis, third_axis)
      axis_parms = self._vells_data.getActiveAxisParms()
      plot_array = self._vells_data.getActiveData().copy()
      if not self._visu_plotter is None:
        self._visu_plotter.reset_zoom()
        self._visu_plotter.delete_cross_sections()
        self._visu_plotter.setAxisParms(axis_parms)
        self._visu_plotter.reset_color_bar(True)
        self._visu_plotter.array_plot(plot_array)
      else:
        if not self.ND_plotter is None:
          self.ND_plotter.delete_vtk_renderer()
          self.ND_plotter.array_plot(plot_array)
          self.ND_plotter.setAxisParms(axis_parms)


  def setArraySelector (self,lcd_number, slider_value, display_string):
    """ callback to handle a request from the N-dimensional
        controller that the user has changed an index into a dimension 
    """
    if self._vells_plot:
      self._vells_data.updateArraySelector(lcd_number,slider_value)
      plot_array = self._vells_data.getActiveData().copy()
      if not self._visu_plotter is None:
        self._visu_plotter.reset_color_bar(True)
        self._visu_plotter.array_plot(plot_array, data_label='data: '+ display_string)
      else:
        if not self.ND_plotter is None:
          self.ND_plotter.array_plot(plot_array, data_label='data: '+ display_string)

  def adjust_selector (self):
    """ instantiate and/or adjust contents of ResultsRange object """
    if self.results_selector is None:
      # used when showing spigot data
      self.results_selector = ResultsRange(self.layout_parent,horizontal=True)
      self.results_selector.setMaxValue(self.max_list_length)
      self.results_selector.set_offset_index(0)
      self.layout.addWidget(self.results_selector, 3,1,Qt.AlignHCenter)
      self.results_selector.show()
      self.results_selector.result_index.connect(self.replay_data)
      self.results_selector.adjust_results_buffer_size.connect(self.set_results_buffer)
      if not self._visu_plotter is None:
        self._visu_plotter.setResultsSelector()
      if self._plot_type == 'realvsimag':
        self._visu_plotter.plot.show_results_selector.connect(self.show_selector)
      else:
        if not self._visu_plotter is None:
          self._visu_plotter.show_results_selector.connect(self.show_selector)
    self.results_selector.set_emit(False)
    self.results_selector.setRange(self.data_list_length-1)
    self.results_selector.setLabel(self.label)
    self.results_selector.set_emit(True)

  def show_selector (self, do_show_selector):
    """ callback to show or hide a ResultsRange object """
    if do_show_selector:
      self.results_selector.show()
    else:
      self.results_selector.hide()

# I'm not sure when/where this function gets called
  def adjust_spectrum_selector (self):
    """ instantiate and/or adjust contents of ResultsRange object """
    if self.spectrum_node_selector is None:
#     self.spectrum_node_selector = ResultsRange(self.layout_parent)
      self.spectrum_node_selector = ResultsRange(self.layout_parent,horizontal=True)
      self.spectrum_node_selector.setStringInfo(' spectrum ')
      self.layout.addWidget(self.spectrum_node_selector, 5, 1, Qt.AlignHCenter)
      self.spectrum_node_selector.show()
      self.spectrum_node_selector.result_index.connect(self.select_spectrum_node)
#     Qt.QObject.connect(self.spectrum_node_selector, Qt.SIGNAL('adjust_results_buffer_size'), self.set_spectrum_node_buffer)
      self._visu_plotter.show_results_selector.connect(self.show_spectrum_selector)
    self.spectrum_node_selector.set_emit(False)
    self.spectrum_node_selector.setMaxValue(len(self.leaf_node_list),False)
    self.spectrum_node_selector.setRange(len(self.leaf_node_list), False)
    self.spectrum_node_selector.setLabel(self.label)
    self.spectrum_node_selector.disableContextmenu()
    self.spectrum_node_selector.set_emit(True)
    

  def show_spectrum_selector (self, do_show_selector):
    """ callback to show or hide a ResultsRange object """
    if do_show_selector:
      self.spectrum_node_selector.show()
    else:
      self.spectrum_node_selector.hide()

  def set_results_buffer (self, result_value):
    """ callback to set the number of results records that can be
        stored in a results history buffer 
    """ 
    if result_value < 0:
      return
    self.max_list_length = result_value
    if len(self.data_list_labels) > self.max_list_length:
      differ = len(self.data_list_labels) - self.max_list_length
      for i in range(differ):
        del self.data_list_labels[0]
        del self.data_list[0]

    if len(self.data_list) != self.data_list_length:
      self.data_list_length = len(self.data_list)

  def ND_controller_showDisplay(self, show_self):
    """ tells ND_Controller to show or hide itself """
    if not self.ND_Controls is None:
      self.ND_Controls.showDisplay(show_self)

  def set_ND_controls (self, labels=None, parms=None, num_axes=2):
    """ this function adds the extra GUI control buttons etc if we are
        displaying data for a numpy array of dimension 3 or greater """

    self.ND_Controls = plot_func.create_ND_Controls(self.layout, self.layout_parent, self.array_shape, self.ND_Controls, self.ND_plotter, labels, parms, num_axes)

    self.ND_Controls.sliderValueChanged.connect(self.setArraySelector)
    self.ND_Controls.defineSelectedAxes.connect(self.setSelectedAxes)

  def show_3D_Display(self, display_flag_3D):
    if not has_vtk:
      return
    if self._vells_data is None:
      Message = '3D displays are not yet implemented for this data type'
      mb_reporter = Qt.QMessageBox.information(None, "ResultPlotter", Message)
      return
    axis_increments = None
    if not display_flag_3D:
      plot_array = self._vells_data.getActiveData().copy()
      if plot_array.min() == plot_array.max():
        return
    axis_increments = self._visu_plotter.getActiveAxesInc()

    self._visu_plotter = plot_func.delete_2D_Plotters(self.colorbar, self._visu_plotter)
    if not self.status_label is None:
      self.status_label.setParent(Qt.QWidget())
      self.status_label = None
    if self.ND_plotter is None:
      self.ND_plotter = plot_func.create_ND_Plotter (self.layout, self.layout_parent)
      self.ND_plotter.show_2D_Display.connect(self.show_2D_Display)
      self.ND_plotter.show_ND_Controller.connect(self.ND_controller_showDisplay)
    else:
      self.ND_plotter.delete_vtk_renderer()
      self.ND_plotter.show_vtk_controls()

# create 3-D Controller
    if display_flag_3D:
      self.set_ND_controls(self.ND_labels, self.ND_parms,num_axes=3)

      self._vells_data.set_3D_Display(display_flag_3D)
      self._vells_data.setInitialSelectedAxes(self.array_rank,self.array_shape,reset=True)
    else:
      if not self.ND_Controls is None:
        self.ND_Controls.setParent(Qt.QWidget())
        self.ND_Controls = None
    self.axis_parms = self._vells_data.getActiveAxisParms()
    plot_array = self._vells_data.getActiveData().copy()
    if plot_array.min() == plot_array.max():
      return
    
    self.ND_plotter.array_plot(plot_array)
    self.ND_plotter.setAxisParms(self.axis_parms)
    if not axis_increments is None:
      self.ND_plotter.setAxisIncrements(axis_increments)

  def show_2D_Display(self, display_flag):
    self.create_2D_plotter()
# create 3-D Controller appropriate for 2-D screen displays
    self.set_ND_controls(self.ND_labels, self.ND_parms,num_axes=2)
    self._vells_data.set_3D_Display(False)
    self._vells_data.setInitialSelectedAxes(self.array_rank,self.array_shape, reset=True)
    self._visu_plotter.set_original_array_rank(3)
    self.plot_vells_data(store_rec=False)
    if not self.results_selector is None:
        self._visu_plotter.show_results_selector.connect(self.show_selector)
        self._visu_plotter.setResultsSelector()
        self.results_selector.show()

  def set_ColorBar (self):
    """ this function adds a colorbar for 2-D displays """
    # create two color bars in case we are displaying complex arrays
    self.colorbar = plot_func.create_ColorBar(self.layout, self.layout_parent, self._visu_plotter, self.plotPrinter)

Grid.Services.registerViewer(dmi_type('MeqResult',record),ResultPlotter,priority=10)
#Grid.Services.registerViewer(meqds.NodeClass('MeqDataCollect'),ResultPlotter,priority=10)
#Grid.Services.registerViewer(meqds.NodeClass('MeqDataConcat'),ResultPlotter,priority=10)
Grid.Services.registerViewer(meqds.NodeClass(),ResultPlotter,priority=25)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import Timba
import MeqGUI
from MeqGUI import Grid, GUI, Plugins
from Timba.dmi import *
from Timba.utils import *
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.Grid.Debug import *
from Timba import *
from MeqGUI import GUI, Plugins, Grid

import weakref
import re
import gc
import types

from PyQt4.Qt import *
import Kittens.widgets
from Kittens.widgets import PYSIGNAL
    
# ====== Grid.Page ===========================================================
# manages one page of a gridded workspace
#
class Page (object):
  class GridRow (QSplitter):
    def __init__(self,parent):
      QSplitter.__init__(self,Qt.Horizontal,parent);
      self._cells = [];
    def cells (self):
      return self._cells;
      
  def __init__ (self,gw,parent_widget,max_nx=5,max_ny=5,fixed_cells=False):
    self._topgrid = QSplitter(Qt.Vertical,parent_widget);
    self._gw = gw;
    self.max_nx     = max_nx;
    self.max_ny     = max_ny;
    self.max_items  = max_nx*max_ny;
    self._rows      = [];
    self._frag_tag = None;
    self._pagename = self._autoname = '';
    self._icon = QIcon();
    self._icon_cou = False;
    # possible layout formats (nrow,ncol)
    self._layouts = [(1,1)];
    for i in range(2,self.max_nx+1):
      self._layouts += [(i,i-1),(i,i)];
    # create cells matrix
    for irow in range(self.max_ny):
      row = self.GridRow(self._topgrid);
      row.hide();
      self._rows.append(row);
      for icol in range(self.max_nx):
        pos = (self,irow,icol);
        cell = MeqGUI.Grid.Cell(row,pos,fixed_cell=fixed_cells,page=self);
        row._cells.append(cell);
        cell._clear_slot = curry(self._clear_cell,cell);
        QWidget.connect(cell.wtop(),PYSIGNAL("closed()"),cell._clear_slot);
        cell._drop_slot = curry(self.drop_cell_item,cell);
        QWidget.connect(cell.wtop(),PYSIGNAL("itemDropped()"),
                        cell._drop_slot);
        cell._cv_slot = curry(self.change_viewer,cell);                
        QWidget.connect(cell.wtop(),PYSIGNAL("changeViewer()"),cell._cv_slot);
        # cell._display_slot = curry(self._display_data_item,parent=weakref.ref(cell));
        # QWidget.connect(cell.wtop(),PYSIGNAL("displayDataItem()"),
        #                cell._display_slot);
    # prepare layout
    self.set_layout(0);
    
  def set_icon (self,icon,clear_on_update=False):
    _dprint(2,icon);
    self._icon = icon or QIcon();
    self._icon_cou = clear_on_update;
    self.wtop().emit(SIGNAL("setIcon"),self._icon,);
    
  def clear_icon (self):
    self.set_icon(None);
    
  def clear_name (self):
    self.set_name(self._autoname);
    
  def get_icon (self):
    return self._icon;
    
  def set_name (self,name,auto=False):
    _dprint(2,name,auto);
    self._pagename = name;
    if auto:
      self._autoname = name;
    self.wtop().emit(SIGNAL("setName"),name,);
  
  def get_name (self):
    return self._pagename;
    
  def is_auto_name (self):
    return self._autoname == self._pagename;
    
  def gw (self):
    """returns parent Grid.Workspace object""";
    return self._gw;
    
  def set_fragile_tag (self,tag=None):
    """A "fragile tag" disappears whenever a page is updated with anything.
    This is useful for, e.g., keeping track of whether a page has been loaded
    from a pagemark and not modified since, etc.""";
    if tag is None:
      self._frag_tag = None;
    else:
      self._frag_tag = weakref.ref(tag);
      
  def get_fragile_tag (self):
    if self._frag_tag is None:
      return None;
    return self._frag_tag();
      
  def updated (self):
    _dprint(3,'page updated');
    self._frag_tag = None;
    if self._icon_cou:
      self.clear_icon();
    self.wtop().emit(SIGNAL("updated"));
    
  def change_viewer (self,cell,dataitem,viewer):
    self.updated();
    dataitem.cleanup();
    cell.wipe();
    MeqGUI.Grid.Services.addDataItem(dataitem,viewer=viewer,
                  gw=self._gw,position=cell.grid_position());
                  
  def drop_cell_item (self,cell,item,event):
    source_cell = getattr(event.source(),'_grid_cell',None);
    if not isinstance(source_cell,MeqGUI.Grid.Cell):
      source_cell = None;
    our_content = cell.content_dataitem();
    # figure out if we need to swap source/destination cells
    action = None;
    # dropping item from another cell: show menu for move/swap
    if source_cell:
      try: menu = self._cell_swap_menu;
      except AttributeError:
        menu = self._cell_swap_menu = QMenu(self.wtop());
        self._menu_slab = Kittens.widgets.addMenuLabel(menu,"");
        self._qa_move_cell   = menu.addAction(pixmaps.slick_redo.icon(),"Move cell");
        self._qa_dup_cell    = menu.addAction(pixmaps.slick_editcopy.icon(),"Duplicate cell");
        self._qa_swap_cells  = menu.addAction(pixmaps.slick_reload.icon(),"Swap cells");
        self._qa_cancel_drop = menu.addAction(pixmaps.red_round_cross.icon(),"Cancel drop");
      self._menu_slab.setText("<nobr>dragging "+item.caption+"</nobr>");
      self._qa_swap_cells.setVisible(our_content is not None); 
      pos = cell.wtop().mapToGlobal(event.pos());
      action = menu.exec_(pos);
      # cancel selected or menu closed, do nothing
      if not action or action is self._qa_cancel_drop:
        return;
    else:
      # assign 1,2,3 to these if menu is not createed (for comparisons below)
      self._qa_move_cell,self._qa_dup_cell,self._qa_swap_cells = 1,2,3;
    # clear our own cell
    if our_content:
      our_content.cleanup();
      cell.wipe();
    # if duplicate cell is selected, create a new data item
    if action is self._qa_dup_cell:
      item = MeqGUI.Grid.DataItem(item);
    # move cell is selected, clear source cell completely
    elif action is self._qa_move_cell:
      if source_cell:
        source_cell.close();
    # swap is selected, move our content to source cell
    elif action is self._qa_swap_cells:
      item.cleanup();
      source_cell.wipe();
      source_pos = source_cell.grid_position();
      MeqGUI.Grid.Services.addDataItem(our_content,
          gw=source_pos[0].gw(),position=source_pos);
    # add content for this cell
    MeqGUI.Grid.Services.addDataItem(item,gw=self.gw(),position=cell.grid_position());
  
  def num_layouts (self):
    return len(self._layouts);
    
  def current_layout (self):
    return (self._cur_layout_num,) + self._cur_layout;
    
  def has_content (self):
    """returns True if page has content on it""";
    (nrow,ncol) = self._cur_layout;
    for row in self._rows[:nrow]:
      for cell in row._cells[:ncol]:
        if cell.content_udi():
          return True;
    return False;
  
  def get_cell (self,row,col):
    return self._rows[row]._cells[col];
  
  # changes current layout scheme
  def set_layout (self,nrow,ncol=None):
    """set_layout(n) sets layout #n. set_layout(nr,nc) sets minimal layout 
    to display nr x nc cells""";
    self.updated();
    if ncol is None:
      nlo = nrow;
    else:
      for (i,(nr,nc)) in enumerate(self._layouts):
        if nr >= nrow and nc >= ncol:
          nlo = i;
          break;
    (nrow,ncol) = self._cur_layout = self._layouts[nlo];
    self._cur_layout_num = nlo;
    _dprint(4,"setting layout:",self._cur_layout);
    for row in self._rows[:nrow]:
      for cell in row.cells()[:ncol]: 
        # if not cell.is_empty(): 
        cell.show();
      for cell in row.cells()[ncol:]: cell.hide();
      row.show();
    for row in self._rows[nrow:]:
      row.hide();
    self.align_layout();
    # emit signal for change of layout
    self.wtop().emit(SIGNAL("layoutChanged"),nlo,len(self._layouts),nrow,ncol);
    return self._cur_layout;
    
  # increments current layout scheme by one (i.e. adds more windows)
  def next_layout (self):
    try: return self.set_layout(self._cur_layout_num+1);
    except IndexError: 
      return None;
      
  def align_layout (self):
    xsizes = [1000]*self.max_nx;
    list(map(lambda row:row.setSizes(xsizes),self._rows));
    self._topgrid.setSizes([1000]*self.max_ny);
    
  # returns top-level widget
  def wtop   (self):
    return self._topgrid;
    
  def clear (self):
    _dprint(2,'GriddedPage: clearing');
    self.updated();
    self.set_layout(0);
    for row in self._rows:
      _dprint(2,'GriddedPage: clearing row',row);
      list(map(lambda cell:cell.close(),row.cells()));
    self.clear_icon();
    self.clear_name();
    
  def find_cells (self,udi,new=False,avoid_pos=None,nrow=1,ncol=1):
    """Finds a free cell if one is available, switches to the next layout
    as needed. Returns Cell object, or None if everything is full.
    If new=False, tries to reuse unpinned cells before switching layouts.
    When picking a cell to reuse, avoids the following cells:
      * pinned cells
      * cells with a leader (so that blocks of cells are never broken up)
      * cell at position avoid_pos
        (this option is specified when a viewer in a cell is used to launch
        another viewer, e.g. for a sub-data item -- in this case the "child"
        viewer shouldn't reuse the cell of the parent).
      * cells whose contents are "parents" of this one
        (i.e. whose UDI is a prefix of this UDI) 
    If new=True, always allocates a new cell.
    """;
    _dprint(2,udi);
    (avx,avy) = avoid_pos or (-1,-1);
    # loop over layouts until we find a cell (or run out of layouts)
    while True:
      (nrow,ncol) = self._cur_layout;
      # find free space in layout
      for icol in range(ncol):
        for (irow,row) in enumerate(self._rows[:nrow]):
          if icol != avx or irow != avy:
            cell = row.cells()[icol];
            if cell.is_empty():
              return (cell,);
      # no free space, try to find an unpinned cell (starting from the back)
      if not new:
        for icol in range(ncol-1,-1,-1):
          for irow in range(nrow-1,-1,-1):
            cell = self._rows[irow].cells()[icol];
            # avoid: pinned cells, follower cells (leader not None),
            # parents of our udi, and cells speciified by the avoid position
            _dprint(5,cell.is_pinned(),cell.leader(),icol,irow);
            if ( not cell.is_pinned() ) and          \
               ( cell.leader() is None ) and     \
               ( not ( icol == avx and irow == avy) ) and \
               ( not cell.is_parent_of(udi) ):
              cell.wipe();
              return (cell,);
      # current layout is full: proceed to next layout
      if not self.next_layout():
        return None;
        
  def alloc_cells (self,pos,nrow=1,ncol=1):
    self.updated();
    (row,col) = pos;
    _dprint(2,row,col);
    # increment layout until we can house the cell position
    while col >= self._cur_layout[1] or row >= self._cur_layout[0]:
      if not self.next_layout():
        return None;
    row = self._rows[row];
    cell = row.cells()[col];
    if not cell.is_empty():
      cell.wipe();
    return (cell,);

  def _clear_cell (self,cell):
    # if a cell is closed and layout is not empty, rearrange the layout
    if self._cur_layout_num:
      self.rearrange_cells();
      
  # rearranges cells by getting rid of empty rows and columns
  def rearrange_cells (self):
    nrow = 0;
    ncol = 0;
    # find max dimensions of non-empty cells
    for (irow,row) in enumerate(self._rows):
      for (icol,cell) in enumerate(row.cells()):
        if not cell.is_empty():
          nrow = max(nrow,irow);
          ncol = max(ncol,icol);
    nrow += 1;
    ncol += 1;
    # are they good for the current layout?
    if nrow == self._cur_layout[0] and ncol == self._cur_layout[1]:
      return;
    # find suitable layout
    for (i,(nr,nc)) in enumerate(self._layouts):
      if nr >= nrow and nc >= ncol:
        self.set_layout(i);
        break;
    else:
      raise Grid.Error("failed to find a suitable layout");




#!/usr/bin/env python3

# a class to generate control buttons etc for use in controlling N-dimensional
# displays


#% $Id: ND_Controller.py 5418 2007-07-19 16:49:13Z oms $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#

# modules that are imported
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from qwt.qt.QtGui import (QApplication, QGridLayout,QHBoxLayout,
         QLabel, QSizePolicy, QSlider, QPushButton, QVBoxLayout, QSpinBox, QSpacerItem)
from qwt.qt.QtGui import QPen, QColor,QWidget, QImage, qRgba
from qwt.qt.QtCore import Qt, QObject, pyqtSignal


# the AxisRange class is directly adapted from the Qt/PyQt 
# tutorial code examples

class AxisRange(QWidget):
    """ a spinbox and a slider, either of which can be used to specify
        a value from within an allowed range
    """
    axis_number = pyqtSignal('int')
    Value_Changed = pyqtSignal(int, int, str)

    def __init__(self, ax_number=1, axis_parms=None,parent=None, name=""):
        """ specify the layout of the spinbox and the slider """
        QWidget.__init__(self, parent)

        self.button = QPushButton(' ', self)
        self.ax_number = ax_number
        self.is_on = False
        self.axis_parms=axis_parms
        self.button_label = None

        self.spinbox = QSpinBox(self)
        self.spinbox.setMinimum(0)
        self.spinbox.setMaximum(99)
        self.spinbox.setWrapping(True)
        self.maxVal = 99

        self.slider = QSlider(Qt.Horizontal, self)
        self.slider.setTickPosition(QSlider.TicksBelow)
        self.slider.setTickInterval(10)
        self.slider.setRange(0, 99)
        self.maxVal = 99
        self.active = False

        self.label_info = QLabel(' ', self)

        self.resetValue()

        self.button.clicked.connect(self.emit_axis_number)
        self.slider.valueChanged.connect(self.update_slider)
        self.spinbox.valueChanged.connect(self.update_spinbox)

        self.layout = QGridLayout(self)
        self.layout.addWidget(self.label_info,0,0)
        self.layout.addWidget(self.button,1,0)
        self.layout.addWidget(self.spinbox, 0,2)
        self.layout.addWidget(self.slider, 1,2)

    def emit_axis_number(self):
#     print 'emitting signal ', self.ax_number
      self.axis_number.emit(self.ax_number)

    def isOn(self):
      return self.is_on

    def setOn(self,value):
      self.is_on = value

    def getButton(self):
      return self.button

    def setLabel(self, label):
        """ set label on a button """
        self.button_label = label
        self.button.setText(label)

    def value(self):
        """ return current value of the slider """
        return self.slider.value()

    def hide_display(self):
        self.spinbox.hide()
        self.slider.hide()
        self.label_info.hide()

    def show_display(self):
        self.spinbox.show()
        self.slider.show()
        self.label_info.show()


    def resetValue(self):
        """ resets displayed values to zero """
        display_str = None
        if self.axis_parms is None or not self.active:
          display_str = ''
        else:
          delta_vells = (self.axis_parms[1] - self.axis_parms[0]) / self.maxVal
          index = self.axis_parms[0] + 0.5 * delta_vells 
          if abs(index) < 0.0000001:
            index = 0.0
          display_str = "%-.4g" % index
        self.slider.setValue(0)
        self.spinbox.setValue(0)
        self.label_info.setText(' ' + display_str)

    def setDisplayString(self, value ):
      """ define current display string from Vells parameters """
      display_str = ''
      if not self.axis_parms is None:
        delta_vells = (self.axis_parms[1] - self.axis_parms[0]) / self.maxVal
        index = self.axis_parms[0] + (value + 0.5) * delta_vells 
        if abs(index) < 0.0000001:
          index = 0.0
        display_str = "%-.4g" % index
        self.label_info.setText(' ' + display_str)
      if not self.button_label is None:
        display_str = self.button_label + ' ' + display_str
      
      print('parms', self.ax_number, value, display_str)
      self.Value_Changed.emit(self.ax_number, value, display_str)

    def setRange(self, array_shape):
        """ make sure slider and spinbox both have same allowable range """
        self.slider.setRange(0, array_shape-1)
        self.spinbox.setMaximum(array_shape-1)
        self.maxVal = array_shape

    def setSliderColor(self, color):
        """ set color of slider - red or green """
        self.slider.setStyleSheet("QWidget { background-color: %s }" % color.name())

    def setActive(self, active):
        self.active = active

    def text(self):
        return self.label.text()

    def setText(self, s):
        self.label.setText(s)

    def update_slider(self, slider_value):
        """ synchronize spinbox value to current slider value """
        if self.active:
          self.spinbox.setValue(slider_value)
          self.setDisplayString(slider_value)
        else:
          self.resetValue()

    def update_spinbox(self, spin_value):
        """ synchronize slider value to current spinbox value """
        if self.active:
          self.slider.setValue(spin_value)
          self.setDisplayString(spin_value)
        else:
          self.spinbox.setValue(0)


controller_instructions = \
'''
This control GUI allows you to select a 2 or 3-dimensional sub-array for on-screen display from a larger N-dimensional array. When you select an array for plotting that has 3 or more dimensions, in the case of a <b>2-D</b> display the default start-up plot will show the <b>last two</b> dimensions with the indices into the other dimensions all initialized to zero. <br><br>
So, for example, if we select a 5-d array for display, the last two dimensions (axes 3 and 4 in current notation) do not have sliders associated with them and are only indicated by buttons on the display. The remaining axes have sliders shown in green and indexes for those dimensions initialized to zero. By moving the sliders associated with these axes, you change the indices for the first three dimensions. Alternatively you may change the index associated with a dimension by clicking the spinbox up or down. Note that spinboxes have wrapping turned on: this means that if you have an index with a maximum value of 99, clicking on a spinbox's up arrow will cause the index to wrap around back to zero. You may also jump to a given dimension index by typing the value of that index in the spinbox. Note that if you are displaying a Vellset, the first value displayed in a spinbox is the value of the Vells at the given index (the second number - separated from the first one by blanks).  <br><br> 
You can change the two axes you wish to see displayed on the screen by clicking on any two of the pushbuttons. The sliders associated with these two pushbuttons will then disappear. Then the other axes will have their sliders shown in green - you can move the sliders (and set spinbox contents) to change the array indices for these dimensions.<br><br>
If one is working with a <b>3-D</b> display, we can put up 3 dimensions on the screen at the same time, so the default start up plot will show the last three dimensions (and thus no sliders for them). Also, in this case, to change the axes you wish to see displayed, you must click on any three of the pushbuttons before the display will be modified.<br<br>

'''


class ND_Controller(QWidget):
    defineSelectedAxes = pyqtSignal(int,int,int)
    sliderValueChanged = pyqtSignal(int, int, str)

    def __init__(self, array_shape=None, axis_label=None, axis_parms = None, num_axes = 2, parent=None, name=""):
      QWidget.__init__(self, parent)
# set default number of selectable axes to use 2-D QWT-based display
      self.selectable_axes = num_axes
# create grid layout
      self.layout = QGridLayout(self)
      self.setWhatsThis(controller_instructions)
      self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Minimum)
      self.construct_selectors(array_shape,axis_label,axis_parms)
    # __init__()

    def construct_selectors(self,array_shape=None,axis_label=None,axis_parms=None):
      """ construct a group of AxisRange objects. These objects allow
          the user to specify which dimensions of an N-dimensional
          array are viewable in a 2 or 3-D display
      """
      self.green = QColor(0,0,0)
      self.green.setGreen(255)
      self.red = QColor(0,0,0)
      self.red.setRed(255)
     
      self.axis_parms = axis_parms

# add control buttons and AxisRange selectors
      self.buttons = []
      self.button_number = []
      self.axis_controllers = []
      
      if array_shape is None:
        array_shape = []
        for i in range(len(axis_label)):
          array_shape.append(axis_parms[axis_label[i]][3])
      self.rank = 0
      for i in range(len(array_shape)):
        if array_shape[i] > 1:
          self.rank = self.rank + 1
      self.active_axes = {}
      self.num_selectors = -1
      row = 0
      col = 0
      for i in range(len(array_shape)):
        if array_shape[i] > 1:
          self.num_selectors = self.num_selectors + 1
# add buttons
          button_label = None
          if not axis_label == None:
            button_label = axis_label[i]
          else:
            button_label = 'axis ' + str(i)
          if self.axis_parms is None:
            parms = None
          else:
            if axis_label[i] in self.axis_parms:
              parms = self.axis_parms[axis_label[i]]
            else:
              parms = None
          self.axis_controllers.append(AxisRange(ax_number=self.num_selectors, axis_parms=parms, parent=self))
          self.axis_controllers[self.num_selectors].setLabel(button_label)
          self.axis_controllers[self.num_selectors].setRange(array_shape[i])
          if self.num_selectors <= self.rank - (self.selectable_axes + 1):
            self.axis_controllers[self.num_selectors].setSliderColor(self.green)
            self.axis_controllers[self.num_selectors].setActive(True)
            self.axis_controllers[self.num_selectors].resetValue()
            self.axis_controllers[self.num_selectors].show_display()
          else:
            self.axis_controllers[self.num_selectors].setSliderColor(self.red)
            self.axis_controllers[self.num_selectors].hide_display()

          self.axis_controllers[self.num_selectors].axis_number.connect(self.defineAxes)
          self.axis_controllers[self.num_selectors].Value_Changed.connect(self.update)
          if col == 0:
            spacer = QSpacerItem(22,9, QSizePolicy.Expanding, QSizePolicy.Minimum)
            self.layout.addItem(spacer, row,col)
            col = col + 1
          self.layout.addWidget(self.axis_controllers[self.num_selectors],row,col)
          self.buttons.append(self.axis_controllers[self.num_selectors].getButton())
          self.button_number.append(i)
#         self.buttons[self.num_selectors].setToggleButton(True)
          if self.num_selectors <= self.rank - (self.selectable_axes + 1):
            self.axis_controllers[self.num_selectors].setOn(False)
          else:
            self.axis_controllers[self.num_selectors].setOn(True)
            self.active_axes[self.num_selectors] = True
          if col >= 4:
            col = 0
            row = row + 1
          else:
            col = col + 1

# add one to get number of active selector buttons
      self.num_selectors = self.num_selectors + 1

    def showDisplay(self, show_self):
        if show_self > 0:
          self.show()
        else:
          self.hide()
    # showDisplay

    def set_num_selectable_axes(self, num_axes=2):
        self.selectable_axes = num_axes
        self.redefineAxes()

    def get_num_selectors(self):
        """ gets number of AxisRange objects in the Controller """
        return self.num_selectors

    def defineAxes(self, button_id, do_on=False):
        """ When a button is pressed, this function figures out if
            the user has selected the required number of dimensions
            for extraction. The AxisRange objects for those
            dimensions are colored red.  All data for the selected
            dimensions will be displayed. The remaining AxisRange
            objects are colored green - they enable the user to
            select a single value from their associated dimension.
        """
#       print 'defineAxes button id = ', button_id
        if not self.active_axes is None and len(self.active_axes) == self.selectable_axes:
          self.resetAxes()
        self.axis_controllers[button_id].setOn(True)
#       if do_on:
#         self.axis_controllers[button_id].setOn(True)
        if self.axis_controllers[button_id].isOn():
          self.axis_controllers[button_id].setSliderColor(self.red)
#         self.axis_controllers[button_id].hide_display()
          self.active_axes[button_id] = True
          if len(self.active_axes) == self.selectable_axes:
            first_axis = None
            second_axis = None
            third_axis = None
            for i in range(self.num_selectors):
              if i in self.active_axes:
                if first_axis is None:
                  first_axis = self.button_number[i]
                elif second_axis is None:
                  second_axis = self.button_number[i]
                else:
                  if self.selectable_axes == 3:
                    if third_axis is None:
                      third_axis = self.button_number[i]
                self.axis_controllers[i].setSliderColor(self.red)
                self.axis_controllers[i].setActive(False)
                self.axis_controllers[i].hide_display()
              else:
                self.axis_controllers[i].setSliderColor(self.green)
                self.axis_controllers[i].setActive(True)
                self.axis_controllers[i].show_display()
              self.axis_controllers[i].resetValue()
            self.defineSelectedAxes.emit(first_axis, second_axis,third_axis)
        else:
          if button_id in self.active_axes:
            del self.active_axes[button_id]
          self.axis_controllers[button_id].setSliderColor(self.red)
#         self.axis_controllers[button_id].hide_display()
    # defineAxes

    def redefineAxes(self):
        self.defineAxes(self.num_selectors-1, True)
        self.defineAxes(self.num_selectors-2, True)
    # redefineAxes

    def resetAxes(self):
        """ resets all AxisRange objects to be inactive """
        for i in range(self.num_selectors):
          self.axis_controllers[i].setOn(False)
          self.axis_controllers[i].setSliderColor(self.red)
          self.axis_controllers[i].setActive(False)
#         self.axis_controllers[i].hide_display()
          self.axis_controllers[i].resetValue()
        self.active_axes = {}
    # resetAxes

    def update(self, axis_number, slider_value, display_string):
      """ emits a signal to the data selection program giving
          the new value of an active dimension index 
      """
      print('Caught ', axis_number, slider_value, display_string)
      if not axis_number in self.active_axes:
        if not self.axis_controllers[axis_number].isOn():
          self.sliderValueChanged.emit(self.button_number[axis_number], slider_value, display_string)
        else:
          self.axis_controllers[axis_number].resetValue()
      else:
        self.axis_controllers[axis_number].resetValue()

# class ND_Controller

# the following tests the ND_Controller class
def make():
    axes = (3,4,5,6,7,8,9)
    demo = ND_Controller(axes)
#   axes = (3,4,5)
#   demo = ND_Controller(axes, num_axes=0)
    demo.show()
    return demo

# make()

def main(args):
    app = QApplication(args)
    demo = make()
    app.exec_()

# main()

# Admire
if __name__ == '__main__':
    main(sys.argv)







#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.dmi import *
from Timba.Meq import meqds
from MeqGUI import Grid
from MeqGUI.GUI import browsers
from MeqGUI.GUI import app_proxy_gui

import os
import Timba.Apps

_dbg = verbosity(0,name='meqgui');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

default_state_open =  ({'cache':({'result':({'vellsets':({'0':None},None)},None)},None), \
                         'request':None },None);

defaultResultViewopts = { \
  browsers.RecordBrowser: { 'default_open': default_state_open }, \
};

defaultNodeViewopts = { \
  browsers.RecordBrowser: { 'default_open': default_state_open },
#  NodeBrowser:   { 'default_open': ({'state':_default_state_open},None) } 
};


_patt_Udi_NodeState = re.compile("^/[^/]+/(([^#/]+)|(#[0-9]+))(/.*)?$");

def isBookmarkable (udi):
  "Returns True if udi refers to a valid bookmark-able data item";
  ff = udi.split('/',3);
  if ff[0] != '':
    return False;
  cat = ff[1];
  if cat == 'forest':
    return True;
  elif cat == 'node' and len(ff) > 2 and ff[2]:  # node must have name
    return True;
  return False;

def makeDataItem (udi,data=None,viewer=None,publish=False,viewopts={}):
  """Creates a data item given a UDI""";
  # parse udi
  (cat,name,trailer) = meqds.parse_udi(udi);
  if cat == 'node':
    if not name:
      raise ValueError("invalid UDI: "+udi);
    # check name or node index
    nn = name.split('#',1);
    if not nn[0]:
      name = int(nn[1]);     # use node index if no name given
    node = meqds.nodelist[name];
    # enable publishing but do not ask for a state update -- this will
    # be done by the item later
    if publish:
      meqds.enable_node_publish(node,True,get_state=False);
    if not trailer:
      return makeNodeDataItem(node,viewer,viewopts);
    else:
      (name,caption) = meqds.make_udi_node_caption(node,trailer);
      desc = "node %s#%d, state field %s" % (node.name,node.nodeindex,trailer);
      return Grid.DataItem(udi,
                name=name,caption=caption,desc=desc,
                data=data,
                refresh=curry(meqds.request_node_state,node.nodeindex),
                              viewer=viewer,viewopts=viewopts);
  elif cat == 'forest':
    if not trailer:
      return makeForestDataItem(data,viewer,viewopts);
    else:
      (name,caption) = meqds.make_parsed_udi_caption(cat,name,trailer);
      return Grid.DataItem(udi,
         name=name,caption=caption,desc="Forest state field "+trailer,
         data=data,
         refresh=meqds.request_forest_state,
         viewer=viewer,viewopts=viewopts);
  else:
    raise ValueError("can't display "+udi);

def makeNodeDataItem (node,viewer=None,viewopts={}):
  """creates a GridDataItem for a node""";
  udi = meqds.node_udi(node);
  nodeclass = meqds.NodeClass(node);
  vo = viewopts.copy();
  vo.update(defaultNodeViewopts);
  (name,caption) = meqds.make_udi_node_caption(node,None);
  # curry is used to create a call for refreshing its state
  return Grid.DataItem(udi,name=name,caption=caption,desc=name,
            datatype=nodeclass,
            refresh=curry(meqds.request_node_state,node.nodeindex),
            viewer=viewer,viewopts=vo);

def makeForestDataItem (data=None,viewer=None,viewopts={}):
  """creates a GridDataItem for forest state""";
  data = data or meqds.get_forest_state();
  udi = '/forest';
  (name,caption) = meqds.make_parsed_udi_caption('forest',None,None);
  return Grid.DataItem('/forest',
     name=name,caption=caption,desc='State of forest',
     data=meqds.get_forest_state(),
     refresh=meqds.request_forest_state,viewer=viewer,viewopts=viewopts);

def start_kernel (pathname,args=''):
  _dprint(0,pathname,args);
  app_proxy_gui.gui.log_message(' '.join(('starting kernel process:',pathname,args)));
  pid = Timba.Apps.spawnvp_nowait(pathname,[pathname]+args.split(' '));



#!/usr/bin/env python3

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from PyQt4 import Qt
import sys

# dummy thread object represents main thread, plus all non-qt threads
class _DummyThread (object):
  def __init__(self,name):
    self.name = name;
  def getName (self):
    return self.name;

# map of running threads
_main_thread = _DummyThread("main");
_active_threads = {};

def currentThread ():
  return _active_threads.get(Qt.QThread.currentThread(),_main_thread);

def trace_lines (frame, event, arg):
  if event == "line":
    print(("%s %s:%d"%(id(Qt.QThread.currentThread()),frame.f_code.co_filename,frame.f_lineno)));
  return trace_lines;


class Thread (Qt.QThread):
  def __init__(self,target=None,name=None):
    Qt.QThread.__init__(self);
    self._name = name;
    if not callable(target):
      raise TypeError("Thread target argument must be a callable object");
    self._target = target;
  def run (self):
#    sys.settrace(trace_lines);
    handle = Qt.QThread.currentThread();
    _active_threads[handle] = self;
    if self._name is None:
      self._name = str(handle);
    self._target();
    del _active_threads[handle];
  def getName (self):
    return self._name;
  def setName (self,name):
    self._name = name;
  def join (self):
    return self.wait();
  currentThread = staticmethod(currentThread);

from traceback import print_stack
    
class Lock (object):
  """emulates threading.Lock, by using a QMutex""";
  def __init__(self):
    self._lock = Qt.QMutex(Qt.QMutex.NonRecursive);
    self._locker = None;
  def acquire (self):
##    print id(self),'acquiring in',Qt.QThread.currentThread(),'; locked is ',self._lock.locked();
##    print_stack();
    self._lock.lock();
##    if self._locker is not None:
##      raise RuntimeError,"acquired lock held by %s from %s" % (str(self._locker),str(Qt.QThread.currentThread()));
    self._locker = Qt.QThread.currentThread();
##    print id(self),'acquired in',Qt.QThread.currentThread(),'; locked is ',self._lock.locked();
  def release (self):
##    print id(self),'releasing in',Qt.QThread.currentThread();
##    thr = Qt.QThread.currentThread();
##    if self._locker != thr:
##      raise RuntimeError,"attempt to release lock held by %s from %s" % (str(self._locker),str(thr));
    self._locker = None;
    self._lock.unlock();

class RLock (Lock):
  """emulates threading.RLock, by using a recursive QMutex""";
  def __init__(self):
    self._lock = Qt.QMutex(Qt.QMutex.Recursive);
    self._locker = None;
  
class Condition (Lock):
  """emulates a regular thread.Condition, by using a QMutex and a QCondition""";
  def __init__(self):
    Lock.__init__(self);
    self._cond = Qt.QWaitCondition();
  def wait (self,timeout=None):
    if timeout is None:
      timeout = sys.maxsize;
    else:
      timeout = int(timeout*1000);
##    print id(self),'waiting in ',Qt.QThread.currentThread();
    self._cond.wait(self._lock,timeout);
##    print id(self),'wait over, lock is ',self._lock.locked();
  def notify (self):
##    print id(self),'notify in ',Qt.QThread.currentThread();
    self._cond.wakeOne();
  def notifyAll (self):
##    print id(self),'notifyAll in ',Qt.QThread.currentThread();
    self._cond.wakeAll();

class QThreadWrapper (Qt.QThread):
  def __init__(self,target,name=None,args=(),kwargs={}):
    if not callable(target):
      raise TypeError("target argument must be a callable object");
    Qt.QThread.__init__(self);
    self._target = target;
    self._args = args;
    self._kwargs = kwargs;
    if name is None:
      self._name = str(target);
    else:
      self._name = name;
  def run (self):
    print(("********** QThreadWrapper: running ",self._name));
    self._target(*self._args,**self._kwargs);
  def join (self):
    return self.wait();
